--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Tags;                          use Ada.Tags;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Game;                       use Events.Game;
with Preferences;                       use Preferences;
with Statuses;                          use Statuses;
with Support;                           use Support;

package body Games is

    use Actor_Lists;

    allocate : Allocator := null;

    ----------------------------------------------------------------------------

    -- Deletes the game session's Actors. If 'temporalOnly' is True, only the
    -- temporal actors controlling entities in the attached world will be
    -- deleted. Temporal actors must be deleted when a world is detached from
    -- the game framework. All actors must be deleted at the end of a game
    -- session.
    procedure Delete_Actors( this         : not null access Game'Class;
                             temporalOnly : Boolean := False );

    -- Sets the game session's world to none, deleting it and all temporal
    -- actors currently controlling entities in the world.
    procedure Unset_World( this : not null access Game'Class );

    -- Sets the game session's world to none, deleting all temporal actors
    -- currently controlling entities in the world, and returning the world
    -- object to the caller.
    procedure Unset_World( this  : not null access Game'Class;
                           world : out A_World );

    -- Invokes Pause on all of the actors to stop any on-going actions. The
    -- pausing of an actor is a single action and not a state transition. Actors
    -- should be paused when gameplay is being suspended so that on-going
    -- actions do not persist across the suspension. (For example, press Left,
    -- suspend world, release Left, resume world, on-going walk action causes
    -- the entity to keep walking even though Left is no longer pressed.)
    procedure Pause_Actors( this : not null access Game'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Game return A_Game is
    begin
        if allocate /= null then
            return allocate.all;
        else
            raise Constraint_Error with "Game allocator not registered";
        end if;
    end Create_Game;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Game ) is
    begin
        Limited_Object(this.all).Construct;
        this.corral := Create_Corral( "game logic" );
        this.physics := Create_Physics( this.corral );
        this.pman := Create_Process_Manager( "game",
                                             Constrain( Get_Pref( "application",
                                                                  "max_logic_fps" ),
                                                        0, 120 ), 15 );
        this.sessionVars := Create_Association;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Game ) is
    begin
        this.Delete_Actors;

        if this.world /= null then
            pragma Debug( Dbg( "Game.Delete: Error: " &
                               "A world is still attached to the framework",
                               D_GAME, ERROR ) );
            this.world := null;    -- this is a leak; see the error above
        end if;

        -- Delete game view
        Delete( this.view );

        -- Delete other owned objects
        Delete( this.sessionVars );
        Delete( this.pman );
        Delete( this.physics );
        Delete( this.corral );

        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Event_Listeners( this : access Game ) is
    begin
        this.corral.Add_Listener( A_Event_Listener(this), "End_Game" );
        this.corral.Add_Listener( A_Event_Listener(this), "Load_World" );
        this.corral.Add_Listener( A_Event_Listener(this), "New_Game" );
        this.corral.Add_Listener( A_Event_Listener(this), "Pause_Game" );
        this.corral.Add_Listener( A_Event_Listener(this), "View_Ready" );
    end Add_Event_Listeners;

    ----------------------------------------------------------------------------

    -- package private
    procedure Set_World( this  : not null access Game'Class;
                         world : in out A_World;
                         swap  : Boolean ) is

        ------------------------------------------------------------------------

        procedure Connect_Actors_To_Entities is

            -- Checks for entities to connect to an existing actor (ie: players)
            procedure Connect_Actors( pos : Actor_Lists.Cursor ) is
                actor : constant A_Actor := Element( pos );

                procedure Examine_Entity( e : not null A_Entity ) is
                begin
                    if actor.Set_Entity( e ) then
                        --Dbg( "Session actor " & actor.To_String &
                        --     " controlling entity " & e.To_String );
                        this.actorIndex.Insert( e.Get_Id, actor );
                    end if;
                end Examine_Entity;
            begin
                this.world.Examine_Entities( Examine_Entity'Unrestricted_Access );
            end Connect_Actors;

            --------------------------------------------------------------------

            -- Creates and connects an actor to control an entity, if possible.
            procedure Create_Actors( e : not null A_Entity ) is
                actor : A_Actor;
            begin
                actor := Actors.Create_Actor( External_Tag( e.all'Tag ) );
                if actor /= null then
                    -- we just created this actor with the entities class
                    -- so we can assume that it can control the entity.
                    actor.Set_Entity( e );
                    --Dbg( "Temporal actor " & actor.To_String &
                    --     " controlling entity " & e.To_String );
                    actor.Set_Game( A_Game(this) );
                    this.actorIndex.Insert( e.Get_Id, actor );
                    this.actors.Append( actor );
                end if;
            end Create_Actors;

       begin
            -- Examine each existing (non-temporal) Actor and associate it
            -- with an entity in the world as necessary.
            this.actors.Iterate( Connect_Actors'Unrestricted_Access );

            -- Examine each entity in the world and create a temporal
            -- controlling Actor as necessary.
            this.world.Examine_Entities( Create_Actors'Unrestricted_Access );
        end Connect_Actors_To_Entities;

        ------------------------------------------------------------------------

        oldWorld : A_World := null;
    begin
        if swap then
            -- returns the world after detaching it
            this.Unset_World( oldWorld );
        else
            -- discards the world after detaching it
            this.Unset_World;
        end if;

        this.world := world;
        world := oldWorld;

        if this.world /= null then
            this.world.Attach_To_Framework( A_Game_State(this), this.corral );
            Connect_Actors_To_Entities;
        end if;

        -- The game and physics processes are not attached yet. They will be
        -- attached after the view has notified us that it's ready after the
        -- load, via a View_Ready event.
    end Set_World;

    ----------------------------------------------------------------------------

    procedure Set_World( this  : not null access Game'Class;
                         world : in out A_World ) is
    begin
        this.Set_World( world, swap => False );
    end Set_World;

    ----------------------------------------------------------------------------

    procedure Set_World( this     : not null access Game'Class;
                         newWorld : in out A_World;
                         oldWorld : out A_World ) is
    begin
        this.Set_World( newWorld, swap => True );
        oldWorld := newWorld;
        newWorld := null;
    end Set_World;

    ----------------------------------------------------------------------------

    procedure Delete_Actors( this         : not null access Game'Class;
                             temporalOnly : Boolean := False ) is
        crsr  : Actor_Lists.Cursor := this.actors.First;
        nxt   : Actor_Lists.Cursor;
        actor : A_Actor;
    begin
        this.actorIndex.Clear;
        while Has_Element( crsr ) loop
            nxt := Next( crsr );
            actor := Element( crsr );
            if temporalOnly and then not actor.Is_Temporal then
                -- clear the entities from non-temporal actors if we're only
                -- deleting temporal actors.
                actor.Set_Entity( null );
            else
                -- delete the actor
                this.actors.Delete( crsr );
                Delete( actor );
            end if;
            crsr := nxt;
        end loop;
    end Delete_Actors;

    ----------------------------------------------------------------------------

    not overriding
    procedure End_Game( this : access Game; interrupted : Boolean ) is
    begin
        if this.inProgress then
            if interrupted then
                null;
                pragma Debug( Dbg( "Game: Game session interrupted", D_GAME, INFO ) );
            else
                null;
                pragma Debug( Dbg( "Game: Game session ended", D_GAME, INFO ) );
            end if;

            this.Set_Playing( False );
            this.paused := False;
            this.inProgress := False;

            this.Unset_World;
            this.Delete_Actors;

            Queue_Game_State( playing => False, interrupted => interrupted );
        else
            null;
            pragma Debug( Dbg( "Warning: Can't end game; no game in progress.",
                               D_GAME, WARNING ) );
        end if;
    end End_Game;

    ----------------------------------------------------------------------------

    procedure Game_Var_Add( this : not null access Game'Class;
                            var  : String;
                            val  : Integer ) is
        value : A_Value;
    begin
        this.sessionVars.Set_Value( var, this.sessionVars.As_Integer( var ) + val );
        value := Create_Value( this.sessionVars.As_Integer( var ) );
        Queue_Game_Var_Changed( var, value );
    end Game_Var_Add;

    ----------------------------------------------------------------------------

    function Get_Corral( this : not null access Game'Class ) return A_Corral is
    begin
        return this.corral;
    end Get_Corral;

    ----------------------------------------------------------------------------

    overriding
    function Get_Game_Var( this : access Game; var  : String ) return Integer is
    begin
        return this.sessionVars.As_Integer( var );
    exception
        when others =>
            pragma Debug( Dbg( "Error: Game session variable '" & var &
                               "' not found", D_EVAL, ERROR ) );
            raise;
    end Get_Game_Var;

    ----------------------------------------------------------------------------

    overriding
    function Get_Game_Var( this : access Game; var  : String ) return A_Value is
    begin
        return result : constant A_Value := this.sessionVars.Get_Value( var ) do
            if result = null then
                null;
                pragma Debug( Dbg( "Error: Game session variable '" & var &
                                   "' not found", D_EVAL, ERROR ) );
            end if;
        end return;
    end Get_Game_Var;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Game ) return String is
        pragma Unreferenced( this );
    begin
        return "Game Logic";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Game;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is

        ------------------------------------------------------------------------

        procedure Do_Pause( evt : not null A_Pause_Game_Event ) is
        begin
            if evt.Is_Paused = not this.paused then
                A_Game(this).Pause( not this.paused );
            end if;
        end Do_Pause;

        ------------------------------------------------------------------------

        procedure Do_View_Ready( evt : not null A_View_Ready_Event ) is
        begin
            if evt.Is_Ready then
                this.Set_Playing( True );
            end if;
        end Do_View_Ready;

        ------------------------------------------------------------------------

    begin
        if evt.Get_Id = END_GAME_ID then
            A_Game(this).End_Game( interrupted => True );   -- event from the view
        elsif evt.Get_Id = LOAD_WORLD_ID then
            A_Game(this).Handle_Load_World( A_Load_World_Event(evt), resp );
        elsif evt.Get_Id = NEW_GAME_ID then
            A_Game(this).New_Game;
        elsif evt.Get_Id = PAUSE_GAME_ID then
            Do_Pause( A_Pause_Game_Event(evt) );
        elsif evt.Get_Id = VIEW_READY_ID then
            Do_View_Ready( A_View_Ready_Event(evt) );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Load_World( this : access Game;
                                 evt  : not null A_Load_World_Event;
                                 resp : out Response_Type ) is
        wasPlaying : constant Boolean := this.playing;
        world      : A_World;
    begin
        -- set the playing state to false while we spend time loading resources
        -- the playing state will be set to true later, after the view has
        -- indicated that it is ready for gameplay.
        this.Set_Playing( False );

        -- Loads the world from disk and sets it as the current world. This
        -- This involves sending events to notify the system that a new world
        -- has been loaded. If world can't be loaded, an exception will be
        -- raised by Load_World().
        world := Worlds.Load_World( evt.Get_Filename );
        this.Set_World( world );

        Set_Response( resp, ST_SUCCESS );
    exception
        when e : others =>
            -- Loading the world from disk failed. The current state of the
            -- game world remains unchanged. Restore the previous playing state.
            this.Set_Playing( wasPlaying );
            Set_Response( resp, ST_FAILED, Exception_Message( e ) );
    end Handle_Load_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure New_Game( this : access Game ) is
    begin
        if not this.inProgress then
            pragma Debug( Dbg( "Game.New_Game: Game session begun", D_GAME, INFO ) );
            this.inProgress := True;
            this.paused := False;
            this.Set_Playing( False );
            Queue_Game_State( playing => True );
        else
            null;
            pragma Debug( Dbg( "Game.New_Game: Can't begin game; " &
                               "game already in progress", D_GAME, ERROR ) );
        end if;
    end New_Game;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pause( this : access Game; enabled : Boolean ) is
    begin
        if this.inProgress and then enabled /= this.paused then
            this.paused := enabled;
            this.Set_Playing( not enabled );

            -- Queue an event notifying the system of the new paused state
            Queue_Game_Paused( this.paused );
        end if;
    end Pause;

    ----------------------------------------------------------------------------

    procedure Pause_Actors( this : not null access Game'Class ) is

        procedure Pause_Actor( pos : Actor_Lists.Cursor ) is
        begin
            Element( pos ).Pause;
        end Pause_Actor;

    begin
        this.actors.Iterate( Pause_Actor'Unrestricted_Access );
    end Pause_Actors;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove_Event_Listeners( this : access Game ) is
    begin
        this.corral.Remove_Listener( A_Event_Listener(this), "View_Ready" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Pause_Game" );
        this.corral.Remove_Listener( A_Event_Listener(this), "New_Game" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Load_World" );
        this.corral.Remove_Listener( A_Event_Listener(this), "End_Game" );
    end Remove_Event_Listeners;

    ----------------------------------------------------------------------------

    procedure Set_Game_Var( this : not null access Game'Class;
                            var  : String;
                            val  : Integer ) is
        value : A_Value;
    begin
        pragma Debug( Dbg( "Game.Set_Game_Var: " & var & " = " & Image( val ),
                           D_GAME, INFO ) );
        this.sessionVars.Set_Value( var, val );
        value := Create_Value( this.sessionVars.As_Integer( var ) );
        Queue_Game_Var_Changed( var, value );
    end Set_Game_Var;

    ----------------------------------------------------------------------------

    procedure Set_Game_Var( this : not null access Game'Class;
                            var  : String;
                            val  : Boolean ) is
        value : A_Value;
    begin
        pragma Debug( Dbg( "Game.Set_Game_Var: " & var & " = " & Boolean'Image( val ),
                           D_GAME, INFO ) );
        this.sessionVars.Set_Value( var, val );
        value := Create_Value( this.sessionVars.As_Integer( var ) );
        Queue_Game_Var_Changed( var, value );
    end Set_Game_Var;

    ----------------------------------------------------------------------------

    procedure Set_Game_Var( this : not null access Game'Class;
                            var  : String;
                            val  : String ) is
        value : A_Value;
    begin
        pragma Debug( Dbg( "Game.Set_Game_Var: " & var & " = """ & val & """",
                           D_GAME, INFO ) );
        this.sessionVars.Set_Value( var, val );
        value := Create_Value( this.sessionVars.As_String( var ) );
        Queue_Game_Var_Changed( var, value );
    end Set_Game_Var;

    ----------------------------------------------------------------------------

    procedure Set_Playing( this : not null access Game'Class; playing : Boolean ) is
    begin
        if this.inProgress then
            if playing /= this.playing then
                this.playing := playing;
                if this.playing then
                    this.pman.Attach_Async( A_Process(this) );
                    this.pman.Attach_Async( A_Process(this.physics) );
                    if this.world /= null then
                        this.pman.Attach_Async( A_Process(this.world) );
                    end if;
                else
                    if this.world /= null then
                        this.pman.Detach_Async( A_Process(this.world) );
                    end if;
                    this.pman.Detach_Async( A_Process(this.physics) );
                    this.pman.Detach_Async( A_Process(this) );
                end if;
            end if;
        end if;
    end Set_Playing;

    ----------------------------------------------------------------------------

    procedure Set_View( this : not null access Game'Class;
                        view : in out A_Game_View ) is
    begin
        if this.view = null then
            this.view := view;
            view := null;
        else
            raise Constraint_Error with "View already set";
        end if;
    end Set_View;

    ----------------------------------------------------------------------------

    not overriding
    procedure Start( this : access Game ) is
    begin
        if not this.started and then not this.stopped then
            this.started := True;
            A_Game(this).Add_Event_Listeners;
            this.pman.Attach_Async( A_Process(this.corral) );
            if this.view /= null then
                this.view.Start;
            end if;
            this.pman.Start;
        end if;
    end Start;

    ----------------------------------------------------------------------------

    procedure Stop( this : not null access Game'Class ) is
    begin
        if not this.stopped then
            this.Set_Playing( False );
            this.paused := False;
            this.inProgress := False;
            this.stopped := True;

            this.Unset_World;
            this.Delete_Actors;

            this.pman.Detach_Async( A_Process(this.corral) );
            A_Game(this).Remove_Event_Listeners;

            if this.view /= null then
                this.view.Stop;
            end if;
            this.pman.Stop;
        end if;
    end Stop;

    ----------------------------------------------------------------------------

    procedure Unset_World( this : not null access Game'Class ) is
        world : A_World := null;
    begin
        this.Unset_World( world );
        Delete( world );
    end Unset_World;

    ----------------------------------------------------------------------------

    procedure Unset_World( this  : not null access Game'Class;
                           world : out A_World ) is
    begin
        world := null;
        if this.world /= null then
            this.Set_Playing( False );

            this.Pause_Actors;
            this.world.Detach_From_Framework;
            this.Delete_Actors( temporalOnly => True );

            world := this.world;
            this.world := null;
        end if;
    end Unset_World;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Game ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Games.allocate := allocate;
    end Register_Allocator;

end Games;
