--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Unchecked_Deallocation;
with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.Graphics;                  use Allegro.Graphics;
with Allegro.Mouse;                     use Allegro.Mouse;
with Allegro.Sprites;                   use Allegro.Sprites;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Interfaces;                        use Interfaces;
with Widgets;                           use Widgets;

package body Renderers is

    -- Set by calling Use_Software_Mouse. The application should set this during
    -- Allegro initialization if it determines the mouse must be drawn in
    -- software mode. Ideally this flag should be part of the drawing API.
    using_sw_mouse : Boolean := False;

    ----------------------------------------------------------------------------

    procedure Delete is new Ada.Unchecked_Deallocation( Page_Array, A_Page_Array );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Renderer( win : not null A_Window;
                              fps : Positive ) return A_Renderer is
        this : constant A_Renderer := new Renderer;
    begin
        this.Construct( win, fps );
        return this;
    end Create_Renderer;

    ----------------------------------------------------------------------------

    overriding
    procedure Adjust( this : access Renderer ) is
    begin
        Object(this.all).Adjust;
        this.videoPages := new Page_Array(this.videoPages'Range);
        for i in this.videoPages'Range loop
            this.videoPages(i) := Create_Video_Bitmap(
                                      Get_Width( this.videoPages(i) ),
                                      Get_Height( this.videoPages(i) ) );
            Clear_To_Color( this.videoPages(i), 0 );
        end loop;
        this.drawPage := this.videoPages'First;
        this.clearedPage := this.videoPages'First - 1;
    end Adjust;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Renderer;
                         win  : not null A_Window;
                         fps  : Positive ) is
        sw : constant Integer := Get_Screen_W;
        sh : constant Integer := Get_Screen_H;
    begin
        Object(this.all).Construct;
        this.frameDelta := Milliseconds( 1000 / fps );
        this.lastFrame := Clock - this.frameDelta;
        this.win := win;
        this.win.Set_Offset( sw / 2 - win.Get_Width * win.Get_Scale / 2,
                             sh / 2 - win.Get_Height * win.Get_Scale / 2 );

        -- enable triple buffering
        if (Gfx_Capabilities and GFX_CAN_TRIPLE_BUFFER) = 0 then
            -- try to enable it (this works on some platforms with Allegro 4)
            if Enable_Triple_Buffer = 0 then
                this.tripleBuffered := True;
            end if;
        end if;

        if (Gfx_Capabilities and GFX_CAN_TRIPLE_BUFFER) /= 0 then
            -- triple buffering enabled!
            this.tripleBuffered := True;
        end if;

        if this.tripleBuffered then
            this.videoPages := new Page_Array(0..2);
            pragma Debug( Dbg( "Renderer: Triple buffering enabled", D_INIT, INFO ) );
        else
            this.videoPages := new Page_Array(0..1);
        end if;

        this.drawPage := this.videoPages'First;
        this.clearedPage := this.videoPages'First - 1;
        for i in this.videoPages'Range loop
            this.videoPages(i) := Create_Video_Bitmap( sw, sh );
            Clear_To_Color( this.videoPages(i), 0 );
        end loop;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Renderer ) is
    begin
        if using_sw_mouse then
            Show_Mouse( null );
        end if;
        for i in this.videoPages'Range loop
            Destroy_Bitmap( this.videoPages(i) );
        end loop;
        Delete( this.videoPages );
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Draw_Frame( this  : not null access Renderer'Class;
                          vpage : not null A_Bitmap;
                          dt    : Time_Span ) is
        pragma Unreferenced( dt );
    begin
        this.win.Draw( vpage );
    end Draw_Frame;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Renderer ) return String is
        pragma Unreferenced( this );
    begin
        return "Renderer";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Renderer; time : Tick_Time ) is
        now : constant Ada.Real_Time.Time := Clock;
    begin
        if now - this.frameDelta > this.lastFrame then
            if this.clearedPage < this.drawPage then
                -- in fullscreen modes, this is the color seen outside the
                -- application window if the chosen resolution * scale doesn't
                -- match the screen resolution. only clear each video page once,
                -- ever, because we'll never draw outside the window.
                Clear_To_Color( this.videoPages(this.drawPage), 0 );
                this.clearedPage := this.drawPage;
            end if;

            -- exceptions raised here will be caught by the Process_Manager and
            -- terminate the application.
            this.Draw_Frame( this.videoPages(this.drawPage), time.elapsed );
            if using_sw_mouse then
                Draw_Sprite( this.videoPages(this.drawPage), Get_Mouse_Sprite, Mouse_X, Mouse_Y );
            end if;

            if this.tripleBuffered then
                -- wait for the previous frame to flip (should be done already)
                while Poll_Scroll /= 0 loop
                    null;
                end loop;
                -- queue a page flip request. this will return immediately.
                if Request_Video_Bitmap( this.videoPages(this.drawPage) ) /= 0 then
                    null;
                    pragma Debug( Dbg( "Renderer Error: " &
                                       "Unable to display video bitmap",
                                       D_GUI, ERROR ) );
                end if;
            else
                -- flip the page. this will wait for vsync.
                if Show_Video_Bitmap( this.videoPages(this.drawPage) ) /= 0 then
                    null;
                    pragma Debug( Dbg( "Renderer Error: " &
                                       "Unable to display video bitmap",
                                       D_GUI, ERROR ) );
                end if;
            end if;

            this.drawPage := (this.drawPage + 1) mod this.videoPages'Length;
            this.lastFrame := now;
        end if;
    end Tick;

    ----------------------------------------------------------------------------

    function Copy( src : A_Renderer ) return A_Renderer is
    begin
        return A_Renderer(Copy( A_Object(src) ));
    end Copy;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Renderer ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    --==========================================================================

    procedure Use_Software_Mouse is
    begin
        using_sw_mouse := True;
    end Use_Software_Mouse;

end Renderers;
