--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.Graphics;                  use Allegro.Graphics;
with Allegro.Sprites;                   use Allegro.Sprites;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Scaling;                           use Scaling;
with Support;                           use Support;
with Tanskanen;
with Themes.Palette;                    use Themes.Palette;
with Widgets.Layouts;                   use Widgets.Layouts;

package body Widgets.Containers is

    use Widget_Lists;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Container ) is
    begin
        this.Delete_Children;
        Destroy_Bitmap( this.childbmp );
        Delete( this.childLayout );
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Child( this    : access Container;
                         child   : in out A_Widget;
                         consume : Boolean := True ) is
    begin
        child.Set_Parent( A_Widget(this) );
        this.children.Append( child );
        if consume then
            child := null;
        end if;
    end Add_Child;

    ----------------------------------------------------------------------------

    not overriding
    procedure Apply_Container_Layout( this  : access Container;
                                      child : not null A_Widget ) is
    begin
        if this.childLayout /= null then
            this.childLayout.Apply( child );
        end if;
    end Apply_Container_Layout;

    ----------------------------------------------------------------------------

    not overriding
    procedure Bring_To_Front( this  : access Container;
                              child : not null A_Widget ) is
        pos : Widget_Lists.Cursor := this.children.Find( child );
    begin
        if Has_Element( pos ) then
            this.children.Delete( pos );
            this.children.Append( child );
        else
            raise WIDGET_NOT_FOUND with "Widget is not a child";
        end if;
    end Bring_To_Front;

    ----------------------------------------------------------------------------

    not overriding
    procedure Delete_Child( this  : access Container;
                            child : in out A_Widget ) is
        pos : constant Widget_Lists.Cursor := this.children.Find( child );
    begin
        if Has_Element( pos ) then
            child.Set_Visible( False );
            A_Container(this).Remove_Child( child );
            Delete( child );
        else
            raise WIDGET_NOT_FOUND with "Widget is not a child";
        end if;
    end Delete_Child;

    ----------------------------------------------------------------------------

    not overriding
    procedure Delete_Children( this : access Container ) is

        procedure Examine( pos : Widget_Lists.Cursor ) is
            child : A_Widget := Widget_Lists.Element( pos );
        begin
            child.Set_Visible( False );
            Delete( A_Limited_Object(child) );
        end Examine;

    begin
        this.children.Iterate( Examine'Access );
        this.children.Clear;
    end Delete_Children;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw( this : access Container;
                    bmp  : not null A_Bitmap;
                    x, y : Integer ) is
        do_scale : constant Boolean := this.zoom /= 1.0;
    begin
        if this.visible and then this.bmp /= null then

            -- update content if dirty
            if this.dirty or else this.drawForeground then
                this.dirty := False;

                -- 1. clear to a solid color
                if this.transparent then
                    -- make it transparent
                    Clear_To_Transparent( this.drawbmp );
                else
                    -- make it a solid color background (color doesn't matter)
                    -- but magenta makes it obvious for the developer to see
                    -- widgets that aren't drawing their full content area
                    pragma Debug( Clear_To_Color( this.drawbmp, base_colors(Magenta) ) );
                    null;
                end if;
                -- 2. draw content background
                A_Widget(this).Draw_Content( Create_Drawing_Context( this.drawbmp, this.vx1, this.vy1, Background ) );
            end if;

            -- scale content and children from .childbmp to .bmp
            if do_scale then
                -- 3a. save content to temporary childbmp
                Blit( this.drawbmp, this.childbmp, 0, 0, 0, 0,
                      Get_Width( this.drawbmp ), Get_Height( this.drawbmp ) );
                -- 3b. draw children to temporary childbmp
                A_Container(this).Draw_Children( this.childbmp );
                -- 3c. draw content foreground
                if this.drawForeground then
                    A_Widget(this).Draw_Content( Create_Drawing_Context( this.childbmp, this.vx1, this.vy1, Foreground ) );
                end if;
                -- 3d. scale content and children to bmp
                Scale( this.childbmp, this.bmp,
                       Get_Width( this.childbmp ), Get_Height( this.childbmp ),
                       this.zoom,
                       Themes.Get.Get_Scaling_Filter );
            else
                -- 3a. blit content to bmp before drawing children
                Blit( this.drawbmp, this.bmp, 0, 0, 0, 0,
                    Get_Width( this.drawbmp ), Get_Height( this.drawbmp ) );
                -- 3b. draw children to bmp
                A_Container(this).Draw_Children( this.bmp );
                -- 3c. draw content foreground
                if this.drawForeground then
                    A_Widget(this).Draw_Content( Create_Drawing_Context( this.bmp, this.vx1, this.vy1, Foreground ) );
                end if;
            end if;

            -- 4. blit widget to given bitmap
            if this.transparent then
                if Get_Color_Depth = 32 then
                    -- Fast ASM blending function
                    Tanskanen.AlphaBlend32( this.bmp, bmp, x, y );
                else
                    -- fast drawing; the transparent portion will show up black
                    Draw_Sprite( bmp, this.bmp, x, y );

                    -- slower drawing but the transparency is always right
                    --Set_Alpha_Blender;
                    --Draw_Trans_Sprite( bmp, this.bmp, x, y );
                    --Solid_Mode;
                end if;
            else
                Blit( this.bmp, bmp, 0, 0, x, y, this.x2 - this.x1 + 1, this.y2 - this.y1 + 1 );
            end if;
        end if;
    end Draw;

    ----------------------------------------------------------------------------

    not overriding
    procedure Draw_Children( this : access Container;
                             bmp  : not null A_Bitmap ) is

        procedure Examine( pos : Widget_Lists.Cursor ) is
            child : constant A_Widget := Element( pos );
        begin
            if Intersect( this.vx1, this.vy1, this.vx2, this.vy2,
                          child.x1, child.y1, child.x2, child.y2 )
            then
                child.Draw( bmp, child.x1 - this.vx1, child.y1 - this.vy1 );
            end if;
        end Examine;

    begin
        this.children.Iterate( Examine'Access );
    end Draw_Children;

    ----------------------------------------------------------------------------

    overriding
    procedure Find_Widget_At( this   : access Container;
                              x, y   : Integer;
                              wx, wy : out Integer;
                              found  : out A_Widget ) is
        -- translate viewport coordinates into content coordinates
        cx : constant Integer := this.vx1 + Integer(x / this.zoom) ;
        cy : constant Integer := this.vy1 + Integer(y / this.zoom);

        ------------------------------------------------------------------------

        procedure Examine( pos : Widget_Lists.Cursor ) is
            child : constant A_Widget := Element( pos );
        begin
            if found = null then
                if cx >= child.x1 and then cx <= child.x2 and then
                   cy >= child.y1 and then cy <= child.y2
                then
                    child.Find_Widget_At( cx - child.x1, cy - child.y1, wx, wy, found );
                end if;
            end if;
        end Examine;

        ------------------------------------------------------------------------

    begin
        found := null;

        if this.visible and then this.enabled then
            -- iterate in reverse to check in Z-order front to back
            this.children.Reverse_Iterate( Examine'Access );
        end if;

        if found = null then
            Widget(this.all).Find_Widget_At( x, y, wx, wy, found );
        end if;
    end Find_Widget_At;

    ----------------------------------------------------------------------------

    not overriding
    procedure Give_Focus( this : access Container; target : not null A_Widget ) is
    begin
        if this.parent /= null then
            A_Container(this.parent).Give_Focus( target );
        else
            null;
            pragma Debug( Dbg( "Error: Unable to focus " & target.To_String &
                               " - not rooted", D_GUI, ERROR ) );
        end if;
    end Give_Focus;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Resize( this : access Container ) is
    begin
        Widget(this.all).Handle_Resize;
        Destroy_Bitmap( this.childbmp );
        if this.drawbmp /= null then
            this.childbmp := Create_Bitmap( Get_Width( this.drawbmp ),
                                            Get_Height( this.drawbmp ) );
        end if;
    end Handle_Resize;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Rooted( this : access Container; rooted : Boolean ) is

        procedure Examine( pos : Widget_Lists.Cursor ) is
            child : constant A_Widget := Element( pos );
        begin
            child.Handle_Rooted( rooted );
        end Examine;

    begin
        this.children.Iterate( Examine'Access );
        Widget(this.all).Handle_Rooted( rooted );
    end Handle_Rooted;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Shown( this : access Container; shown : Boolean ) is

        procedure Examine( pos : Widget_Lists.Cursor ) is
            child : constant A_Widget := Element( pos );
        begin
            if child.Is_Visible then
                child.Handle_Shown( shown );
            else
                -- if the child isn't visible then it's not affected by showing
                -- notifications; skip it.
                null;
            end if;
        end Examine;

    begin
        this.children.Iterate( Examine'Access );
        Widget(this.all).Handle_Shown( shown );
    end Handle_Shown;

    ----------------------------------------------------------------------------

    overriding
    procedure Pack( this : access Container ) is

        procedure Examine( pos : Widget_Lists.Cursor ) is
        begin
            Element( pos ).Pack;
        end Examine;

    begin
        if this.Is_Rooted then
            Widget(this.all).Pack;
            this.children.Iterate( Examine'Access );
        end if;
    end Pack;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove_Child( this  : access Container;
                            child : not null A_Widget ) is
        pos : Widget_Lists.Cursor := this.children.Find( child );
    begin
        if Has_Element( pos ) then
            this.children.Delete( pos );
            child.Set_Parent( null );
        else
            raise WIDGET_NOT_FOUND with "Widget is not a child";
        end if;
    end Remove_Child;

    ----------------------------------------------------------------------------

    overriding
    procedure Set_Zoom( this : access Container; zoom : Float ) is
    begin
        Widget(this.all).Set_Zoom( zoom );
        Destroy_Bitmap( this.childbmp );
        if this.drawbmp /= null then
            this.childbmp := Create_Bitmap( Get_Width( this.drawbmp ),
                                            Get_Height( this.drawbmp ) );
        end if;
    end Set_Zoom;

end Widgets.Containers;
