--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Actions;                           use Actions;
with Allegro.Keyboard;                  use Allegro.Keyboard;
--with Debugging;                         use Debugging;
with Events.Application;                use Events.Application;
with Events.Audio;                      use Events.Audio;
with Events.Game;                       use Events.Game;
with Events.Keen;                       use Events.Keen;
with Events.World;                      use Events.World;
with Game_Views;                        use Game_Views;
with Keyboard;                          use Keyboard;
with Preferences;                       use Preferences;
with Resources;
with Scaling;                           use Scaling;
with Themes;                            use Themes;
with Widgets.Buttons;                   use Widgets.Buttons;
with Widgets.Buttons.Simple_Listeners;
with Widgets.Containers.Game_Screens;   use Widgets.Containers.Game_Screens;
with Widgets.Containers.Game_Screens.Game_Play;
 use Widgets.Containers.Game_Screens.Game_Play;
with Widgets.Containers.Game_Screens.Menus.Main;
 use Widgets.Containers.Game_Screens.Menus.Main;
with Widgets.Containers.Game_Screens.Message;
 use Widgets.Containers.Game_Screens.Message;
with Widgets.Containers.Game_Screens.Retry;
 use Widgets.Containers.Game_Screens.Retry;
with Widgets.Containers.Game_Screens.Title;
 use Widgets.Containers.Game_Screens.Title;
with Widgets.Containers.Screen_Managers;
 use Widgets.Containers.Screen_Managers;
with Widgets.Labels;                    use Widgets.Labels;
with Widgets.Simple_Key_Listeners;

pragma Elaborate_All( Game_Views );
pragma Elaborate_All( Preferences );
pragma Elaborate_All( Widgets.Simple_Key_Listeners );
pragma Elaborate_All( Widgets.Buttons.Simple_Listeners );

package body Game_Views.Keen is

    MIN_LOAD_DELAY : constant Time_Span := Seconds( 2 );

    package Button_Listeners is new Widgets.Buttons.Simple_Listeners(Keen_View);
    use Button_Listeners;

    package Key_Listeners is new Widgets.Simple_Key_Listeners(Keen_View);
    use Key_Listeners;

    ----------------------------------------------------------------------------

   -- Called when the user attempts to quit the game via a key shortcut. This
   -- procedure may do something else first, like ask if the game should be
   -- saved, etc. When the game should be quit, this procedure will queue a
   -- Close_Window event.
    procedure Action_Quit( this   : not null access Keen_View'Class;
                           action : A_Key_Action );

    ----------------------------------------------------------------------------

    procedure Handle_Dialog_Event( this : not null access Keen_View'Class;
                                   evt  : not null A_Dialog_Event );

    procedure Handle_Game_State( this : not null access Keen_View'Class;
                                 evt  : not null A_Game_State_Event );

    procedure Handle_Message_Event( this : not null access Keen_View'Class;
                                    evt  : not null A_Message_Event );

    procedure Handle_World_Property_Changed( this : not null access Keen_View'Class;
                                             evt  : not null A_World_Property_Changed_Event );

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    -- The following procedures update the GUI.

    -- Called when loading begins. Stops any music currently being played, and
    -- shows the loading screen.
    procedure On_Loading_Begin( this : not null access Keen_View'Class );

    -- Called when loading ends. Exits the loading screen and begins playing the
    -- world's music.
    procedure On_Loading_End( this : not null access Keen_View'Class );

    -- Creates all of the widgets required for the game view. This should only
    -- be called once as part of Game_View object construction. An exception
    -- will be raised if an error occurs creating the widgets.
    procedure Populate_View( this  : not null access Keen_View'Class;
                             xres,
                             yres  : Positive;
                             scale : Positive );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Game_View( xres, yres : Positive;
                               scale      : Positive ) return A_Game_View is
        this : A_Keen_View := new Keen_View;
    begin
        this.Construct;
        this.Populate_View( xres, yres, scale );
        return A_Game_View(this);
    exception
        when others =>
            Delete( A_Game_View(this) );
            raise;
    end Create_Game_View;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Keen_View ) is
    begin
        Game_View(this.all).Construct;

        -- Preload all sound effects
        Resources.Preload_Resource( "click.wav", "audio" );
        Resources.Preload_Resource( "enter_level.wav", "audio" );
        Resources.Preload_Resource( "exit_level.wav", "audio" );
        Resources.Preload_Resource( "get_1up.wav", "audio" );
        Resources.Preload_Resource( "get_drop.wav", "audio" );
        Resources.Preload_Resource( "get_gem.wav", "audio" );
        Resources.Preload_Resource( "get_gun.wav", "audio" );
        Resources.Preload_Resource( "get_points.wav", "audio" );
        Resources.Preload_Resource( "keen_dead.wav", "audio" );
        Resources.Preload_Resource( "keen_die.wav", "audio" );
        Resources.Preload_Resource( "keen_floor.wav", "audio" );
        Resources.Preload_Resource( "keen_jump.wav", "audio" );
        Resources.Preload_Resource( "keen_pogo.wav", "audio" );
        Resources.Preload_Resource( "keen_shoot.wav", "audio" );
        Resources.Preload_Resource( "menu_close.wav", "audio" );
        Resources.Preload_Resource( "menu_open.wav", "audio" );
        Resources.Preload_Resource( "paddle_hit1.wav", "audio" );
        Resources.Preload_Resource( "paddle_hit2.wav", "audio" );
        Resources.Preload_Resource( "paddle_miss.wav", "audio" );
        Resources.Preload_Resource( "paddle_score.wav", "audio" );
        Resources.Preload_Resource( "paddle_wall.wav", "audio" );
        Resources.Preload_Resource( "shot_hit.wav", "audio" );
    end Construct;

    ----------------------------------------------------------------------------

    procedure Action_Quit( this   : not null access Keen_View'Class;
                           action : A_Key_Action ) is
        pragma Unreferenced( action );
    begin
        this.Quit;
     end Action_Quit;

    ----------------------------------------------------------------------------

    procedure Auto_Pause( this  : not null access Keen_View'Class;
                          pause : Boolean ) is
    begin
        if this.gameInProgress then
            if pause then
                if not this.paused then
                    this.Pause_Game( True, force => True );
                    this.autoPaused := True;
                end if;
            else
                if this.autoPaused then
                    this.Pause_Game( False, force => True );
                    this.autoPaused := False;
                end if;
            end if;
            this.pauseEnabled := not pause;
        end if;
    end Auto_Pause;

    ----------------------------------------------------------------------------

    procedure Handle_Dialog_Event( this : not null access Keen_View'Class;
                                   evt  : not null A_Dialog_Event ) is
        screen : A_Game_Screen;
    begin
        -- we're expecting a retry level dialog with two choices
        if evt.Get_Desc.Get_Name = "Retry" then
            screen := Create_Retry_Screen( this, "scrRetry", evt.Get_Dialog_Id );
            A_Label(this.Get_Widget( "scrRetry:lblRetryLevel" )).Set_Text( evt.Get_Desc.Get_Text );
            A_Button(this.Get_Widget( "scrRetry:btnRetry1" )).Set_Text( evt.Get_Desc.Get_Choice( 1 ) );
            A_Button(this.Get_Widget( "scrRetry:btnRetry2" )).Set_Text( evt.Get_Desc.Get_Choice( 2 ) );
            A_Screen_Manager(this.Get_Widget( "screenManager" )).Add_Screen( screen );
        end if;
    end Handle_Dialog_Event;

    ----------------------------------------------------------------------------

    procedure Handle_Game_State( this : not null access Keen_View'Class;
                                 evt  : not null A_Game_State_Event ) is
        screen : A_Game_Screen;
    begin
        if evt.Is_Playing then
            -- A game has begun - - - - - -
            this.gameInProgress := True;

            -- clear paused state
            this.Get_Widget( "pnlPaused" ).Set_Visible( False );
            this.autoPaused := False;
            this.pauseEnabled := True;
            this.paused := False;

            -- the menu screen was removed when the event to start the game
            -- was queued.
        else
            -- A game has ended - - - - - -
            this.gameInProgress := False;
            Queue_Stop_Music;

            if not evt.Is_Interrupted then
                -- the game ended naturally by Keen losing all his lives, so
                -- put the menu back up. otherwise, the game session was
                -- interrupted by the user to start a new game or load a game,
                -- so we don't want to put the menu back up.
                screen := Create_Main_Menu_Screen( A_Game_View(this), "scrMainMenu",
                                                   gameInProgress => False );
                A_Screen_Manager(this.Get_Widget( "screenManager" )).Add_Screen( screen );
            end if;
        end if;
    end Handle_Game_State;

    ----------------------------------------------------------------------------

    procedure Handle_Message_Event( this : not null access Keen_View'Class;
                                    evt  : not null A_Message_Event ) is
        msg : A_Game_Screen;
    begin
        msg := Create_Message_Screen( A_Game_View(this), "scrMessage", True );
        Message_Screen(msg.all).Set_Text_Align( evt.Get_Text_Align );
        Message_Screen(msg.all).Set_Text( evt.Get_Text );
        Message_Screen(msg.all).Set_Icon( evt.Get_Icon );
        Message_Screen(msg.all).Set_Closable( True );
        Message_Screen(msg.all).Set_Auto_Pause( True );
        Screen_Manager(this.Get_Widget( "screenManager" ).all).Add_Screen( msg );
    exception
        when others =>
            null; --Dbg( "Message already on display" );
    end Handle_Message_Event;

    ----------------------------------------------------------------------------

    procedure Handle_World_Property_Changed( this : not null access Keen_View'Class;
                                             evt  : not null A_World_Property_Changed_Event ) is
    begin
        if evt.Get_Property_Name = "music" then
            this.worldMusic := To_Unbounded_String( evt.Get_Value );
   elsif evt.Get_Property_Name = "introduction" then
            if this.loadingScreen then
                Message_Screen(this.Get_Widget( "scrLoading" ).all).Set_Text( evt.Get_Value );
            end if;

        end if;
    end Handle_World_Property_Changed;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Close_Request( this : access Keen_View ) is
         pragma Unreferenced( this );
    begin
         Queue_Close_Window;
    end Handle_Close_Request;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Keen_View;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        Game_View(this.all).Handle_Event( evt, resp );
        if evt /= null then

            if evt.Get_Id = WORLD_PROPERTY_CHANGED_ID then
                this.Handle_World_Property_Changed( A_World_Property_Changed_Event(evt) );

            elsif evt.Get_Id = GAME_STATE_ID then
                this.Handle_Game_State( A_Game_State_Event(evt) );

            elsif evt.Get_Id = MESSAGE_ID then
                this.Handle_Message_Event( A_Message_Event(evt) );

            elsif evt.Get_Id = DIALOG_ID then
                this.Handle_Dialog_Event( A_Dialog_Event(evt) );

            end if;
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Loading( this : access Keen_View; loading : Boolean ) is
    begin
        this.loading := loading;
        if this.gameInProgress and then this.loading then
            -- the loading screen is shown here when loading starts, but it
            -- isn't hidden here when loading ends. the Tick procedure is
            -- responsible for hiding the screen after loading has finished and
            -- MIN_LOAD_DELAY has elapsed (to give the player time to read the
            -- loading screen.)
            this.On_Loading_Begin;
        end if;
    end Handle_Loading;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Paused( this : access Keen_View; paused : Boolean ) is
    begin
        this.Get_Widget( "pnlPaused" ).Set_Visible( paused );
    end Handle_Paused;

    ----------------------------------------------------------------------------

    function Is_Paused( this : not null access Keen_View'Class ) return Boolean is
    begin
        return this.paused;
    end Is_Paused;

    ----------------------------------------------------------------------------

    procedure New_Game( this : not null access Keen_View'Class ) is
    begin
        if this.gameInProgress then
            Queue_End_Game;
        end if;
        Queue_New_Game;
    end New_Game;

    ----------------------------------------------------------------------------

    procedure On_Loading_Begin( this : not null access Keen_View'Class ) is
        screen : A_Game_Screen;
    begin
        Queue_Stop_Music;

        screen := Create_Message_Screen( this, "scrLoading", transparent => False );
        Message_Screen(screen.all).Set_Text_Align( Align_Right );
        Message_Screen(screen.all).Set_Text( "One moment" );
        Message_Screen(screen.all).Set_Icon( "loading_keen1" );
        A_Screen_Manager(this.Get_Widget( "screenManager" )).Add_Screen( screen );

        -- the loading screen won't hide for at least MIN_LOAD_DELAY time
        this.loadingScreen := True;
        this.loadingScreenComplete := Clock + MIN_LOAD_DELAY;
    end On_Loading_Begin;

    ----------------------------------------------------------------------------

    procedure On_Loading_End( this : not null access Keen_View'Class ) is
    begin
        this.loadingScreen := False;
        A_Game_Screen(this.Get_Widget( "scrLoading" )).Exit_Screen;

        -- start playing music for the next world
        if Length( this.worldMusic ) > 0 then
            Queue_Play_Music( To_String( this.worldMusic ) );
        end if;

        -- the loading screen is done so let the game know the view is ready
        -- to play
        Queue_View_Ready( True );
    end On_Loading_End;

    ----------------------------------------------------------------------------

    procedure Pause_Game( this  : not null access Keen_View'Class;
                          pause : Boolean;
                          force : Boolean := False ) is
    begin
        if this.gameInProgress then
            if pause = not this.paused then
                if this.pauseEnabled or else force then
                    Queue_Pause_Game( pause );
                end if;
            end if;
        end if;
    end Pause_Game;

    ----------------------------------------------------------------------------

    procedure Populate_View( this  : not null access Keen_View'Class;
                             xres,
                             yres  : Positive;
                             scale : Positive ) is
        win           : A_Window;
        screenManager : A_Screen_Manager;
        screen        : A_Game_Screen;
    begin
        win := Create_Window( this, "win",
                              xres, yres, scale,
                              To_Filter( Get_Pref( "scaling_filter" ) ) );
        this.Set_Window( win );
        win.Add_Listener( Listener( KEY_F4, (ALT=>Yes,others=>No), Key_Press, this, Action_Quit'Access ) );

        -- Screen Manager and initial game screens
        -- Title (in front), Main Menu, Gameplay
        screenManager := Create_Screen_Manager( this, "screenManager" );

            screen := Create_Game_Play_Screen( this, "scrGamePlay" );
            screenManager.Add_Screen( screen );

            screen := Create_Main_Menu_Screen( this, "scrMainMenu",
                                               gameInProgress => False );
            screenManager.Add_Screen( screen );

            screen := Create_Title_Screen( this, "scrTitle" );
            screenManager.Add_Screen( screen );

        win.Add_Widget( A_Widget(screenManager) );
    exception
        when others =>
            -- An exception occurred creation a widget. Delete all the tmp
            -- widget pointers to make sure the widgets being built are freed.
            Delete( A_Widget(screen) );
            Delete( A_Widget(screenManager) );
            Delete( A_Widget(win) );
            raise;
    end Populate_View;

    ----------------------------------------------------------------------------

    procedure Quit( this : not null access Keen_View'Class ) is
        pragma Unreferenced( this );
    begin
        Queue_Close_Window;
    end Quit;

    ----------------------------------------------------------------------------

    overriding
    procedure Start_View( this : access Keen_View ) is
    begin
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Dialog" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Game_State" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Message" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Property_Changed" );
    end Start_View;

    ----------------------------------------------------------------------------

    overriding
    procedure Stop_View( this : access Keen_View ) is
    begin
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Property_Changed" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Message" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Game_State" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Dialog" );
    end Stop_View;

    ----------------------------------------------------------------------------

    procedure Tick( this : access Keen_View; time : Tick_Time ) is
    begin
        Game_View(this.all).Tick( time );

        -- handle the loading screen
        if not this.loading then
            -- loading is finished; is it time to take down the loading screen?
            if this.loadingScreen and then Clock > this.loadingScreenComplete then
                -- loading is complete and the minimum delay has elapsed;
                this.On_Loading_End;
            end if;
        end if;
    end Tick;

begin

    Register_Allocator( Create_Game_View'Access );

end Game_Views.Keen;
